from .eventrequest import EventRequest

from .eventresponse import EventResponse

from .eventstream import \
	EventPermissionError, \
	send_event, \
	get_events, \
	get_current_event_id, \
	channel_permission_changed

from .utils import have_channels

from . import urls

if have_channels():
	from . import routing



import json
import threading
import importlib
import six
from django.conf import settings
from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder
from gripcontrol import HttpStreamFormat

try:
	from urllib import quote
except ImportError:
	from urllib.parse import quote

tlocal = threading.local()

def have_channels():
	try:
		from channels.generic.http import AsyncHttpConsumer
		return True
	except ImportError:
		return False

# return dict of (channel, last-id)
def parse_last_event_id(s):
	out = {}
	parts = s.split(',')
	for part in parts:
		channel, last_id = part.split(':')
		out[channel] = last_id
	return out

def make_id(ids):
	id_parts = []
	for channel, id in six.iteritems(ids):
		enc_channel = quote(channel)
		id_parts.append('%s:%s' % (enc_channel, id))
	return ','.join(id_parts)

def build_id_escape(s):
	out = ''
	for c in s:
		if c == '%':
			out += '%%'
		else:
			out += c
	return out

def sse_encode_event(event_type, data, event_id=None, escape=False, json_encode=True):
	if json_encode:
		data_str = json.dumps(data, cls=DjangoJSONEncoder)
	else:
		data_str = data
	if escape:
		event_type = build_id_escape(event_type)
		data_str = build_id_escape(data_str)
	out = 'event: %s\n' % event_type
	if event_id:
		out += 'id: %s\n' % event_id
	out += 'data: %s\n\n' % data_str
	return out

def sse_error_response(condition, text, extra=None):
	if extra is None:
		extra = {}
	data = {'condition': condition, 'text': text}
	for k, v in six.iteritems(extra):
		data[k] = v
	body = sse_encode_event('stream-error', data, event_id='error')
	return HttpResponse(body, content_type='text/event-stream')

def publish_event(channel, event_type, data, pub_id, pub_prev_id,
		skip_user_ids=None, json_encode=True, **publish_kwargs):
	from django_grip import publish

	if skip_user_ids is None:
		skip_user_ids = []

	content_filters = []
	if pub_id:
		event_id = '%I'
		content_filters.append('build-id')
	else:
		event_id = None
	content = sse_encode_event(event_type, data, event_id=event_id, escape=bool(pub_id), json_encode=json_encode)
	meta = {}
	if skip_user_ids:
		meta['skip_users'] = ','.join(skip_user_ids)
	publish(
		'events-%s' % quote(channel),
		HttpStreamFormat(content, content_filters=content_filters),
		id=pub_id,
		prev_id=pub_prev_id,
		meta=meta,
		**publish_kwargs)

def publish_kick(user_id, channel):
	from django_grip import publish

	msg = 'Permission denied to channels: %s' % channel
	data = {'condition': 'forbidden', 'text': msg, 'channels': [channel]}
	content = sse_encode_event('stream-error', data, event_id='error')
	meta = {'require_sub': 'events-%s' % channel}
	publish(
		'user-%s' % user_id,
		HttpStreamFormat(content),
		id='kick-1',
		meta=meta)
	publish(
		'user-%s' % user_id,
		HttpStreamFormat(close=True),
		id='kick-2',
		prev_id='kick-1',
		meta=meta)

def load_class(name):
	at = name.rfind('.')
	if at == -1:
		raise ValueError('class name contains no \'.\'')
	module_name = name[0:at]
	class_name = name[at + 1:]
	return getattr(importlib.import_module(module_name), class_name)()

# load and keep in thread local storage
def get_class(name):
	if not hasattr(tlocal, 'loaded'):
		tlocal.loaded = {}
	c = tlocal.loaded.get(name)
	if c is None:
		c = load_class(name)
		tlocal.loaded[name] = c
	return c

def get_class_from_setting(setting_name, default=None):
	if hasattr(settings, setting_name):
		return get_class(getattr(settings, setting_name))
	elif default:
		return get_class(default)
	else:
		return None

def get_storage():
	return get_class_from_setting('EVENTSTREAM_STORAGE_CLASS')

def get_channelmanager():
	return get_class_from_setting(
		'EVENTSTREAM_CHANNELMANAGER_CLASS',
		'django_eventstream.channelmanager.DefaultChannelManager')

def add_default_headers(headers):
	headers['Cache-Control'] = 'no-cache'
	headers['X-Accel-Buffering'] = 'no'
	augment_cors_headers(headers)

def augment_cors_headers(headers):
	cors_origin = getattr(settings, 'EVENTSTREAM_ALLOW_ORIGIN', '')

	if cors_origin:
		headers['Access-Control-Allow-Origin'] = cors_origin

	allow_credentials = getattr(settings, 'EVENTSTREAM_ALLOW_CREDENTIALS', False)

	if allow_credentials:
		headers['Access-Control-Allow-Credentials'] = 'true'

	allow_headers = getattr(settings, 'EVENTSTREAM_ALLOW_HEADERS', '')

	if allow_headers:
		headers['Access-Control-Allow-Headers'] = allow_headers



class ChannelManagerBase(object):
	def get_channels_for_request(self, request, view_kwargs):
		raise NotImplementedError()

	# user=None for anonymous
	def can_read_channel(self, user, channel):
		raise NotImplementedError()

	# this only has meaning if storage is enabled
	def is_channel_reliable(self, channel):
		raise NotImplementedError()

class DefaultChannelManager(ChannelManagerBase):
	def get_channels_for_request(self, request, view_kwargs):
		# by default, use view keywords, else query params
		if 'format-channels' in view_kwargs:
			out = set()
			for format_channel in view_kwargs['format-channels']:
				out.add(format_channel.format(**view_kwargs))
			return out
		elif 'channels' in view_kwargs:
			return set(view_kwargs['channels'])
		elif 'channel' in view_kwargs:
			return set([view_kwargs['channel']])
		else:
			return set(request.GET.getlist('channel'))

	def can_read_channel(self, user, channel):
		return True

	def is_channel_reliable(self, channel):
		return True



import copy
import time
import jwt
import six
from gripcontrol import Channel
from django.conf import settings
from django.http import HttpResponse
from .utils import sse_encode_event, make_id, build_id_escape

try:
	from urllib import quote
except ImportError:
	from urllib.parse import quote

class EventResponse(object):
	def __init__(self):
		self.channel_items = {}
		self.channel_last_ids = {}
		self.channel_reset = set()
		self.channel_more = set()
		self.is_next = False
		self.is_recover = False
		self.user = None

	def to_http_response(self, http_request):
		last_ids = copy.deepcopy(self.channel_last_ids)
		event_id = make_id(last_ids)

		body = ''

		if not self.is_next:
			body += ':' + (' ' * 2048) + '\n\n'
			body += 'event: stream-open\ndata:\n\n'

		if len(self.channel_reset) > 0:
			body += sse_encode_event(
				'stream-reset',
				{'channels': list(self.channel_reset)},
				event_id=event_id)

		for channel, items in six.iteritems(self.channel_items):
			for item in items:
				last_ids[channel] = item.id
				event_id = make_id(last_ids)
				body += sse_encode_event(
					item.type,
					item.data,
					event_id=event_id,
					escape=True)

		resp = HttpResponse(body, content_type='text/event-stream')

		more = (len(self.channel_more) > 0)

		user_id = str(self.user.id) if self.user else 'anonymous'

		params = http_request.GET.copy()
		params['link'] = 'next'
		if more:
			params['lastEventId'] = event_id
			if 'recover' in params:
				del params['recover']
		else:
			params['recover'] = 'true'
			if 'lastEventId' in params:
				del params['lastEventId']
		es_meta = {
			'iss': 'es',
			'exp': int(time.time()) + 3600,
			'channels': list(self.channel_items.keys()),
			'user': user_id
		}
		params['es-meta'] = six.ensure_text(jwt.encode(es_meta,
				settings.SECRET_KEY.encode('utf-8')))
		next_uri = http_request.path + '?' + params.urlencode()

		instruct = http_request.grip.start_instruct()

		instruct.set_next_link(next_uri)

		for channel in six.iterkeys(self.channel_items):
			enc_channel = quote(channel)
			last_id = last_ids.get(channel)
			gc = Channel('events-%s' % enc_channel, prev_id=last_id)
			if last_id:
				gc.filters.append('build-id')
			gc.filters.append('skip-users')
			instruct.add_channel(gc)

		gc = Channel('user-%s' % user_id)
		gc.filters.append('require-sub')
		instruct.add_channel(gc)

		if not more:
			instruct.set_hold_stream()

		if len(last_ids) > 0:
			id_parts = []
			for channel in six.iterkeys(last_ids):
				enc_channel = quote(channel)
				id_parts.append('%s:%%(events-%s)s' % (build_id_escape(
						enc_channel), enc_channel))
			id_format = ','.join(id_parts)
			instruct.meta['id_format'] = id_format

		instruct.meta['user'] = user_id

		instruct.set_keep_alive('event: keep-alive\ndata:\n\n', 20)

		return resp



import copy
import threading
import asyncio
import six
from django.http import HttpResponseBadRequest
from channels.generic.http import AsyncHttpConsumer
from channels.http import AsgiRequest
from channels.db import database_sync_to_async
from .utils import add_default_headers

MAX_PENDING = 10

class Listener(object):
	def __init__(self):
		self.loop = asyncio.get_event_loop()
		self.aevent = asyncio.Event()
		self.user_id = ''
		self.channels = set()
		self.channel_items = {}
		self.overflow = False
		self.error = ''

	def wake_threadsafe(self):
		self.loop.call_soon_threadsafe(self.aevent.set)

class ListenerManager(object):
	def __init__(self):
		self.lock = threading.Lock()
		self.listeners_by_channel = {}

	def add_listener(self, listener):
		self.lock.acquire()
		try:
			for channel in listener.channels:
				clisteners = self.listeners_by_channel.get(channel)
				if clisteners is None:
					clisteners = set()
					self.listeners_by_channel[channel] = clisteners
				clisteners.add(listener)
		finally:
			self.lock.release()

	def remove_listener(self, listener):
		self.lock.acquire()
		try:
			for channel in listener.channels:
				clisteners = self.listeners_by_channel.get(channel)
				clisteners.remove(listener)
				if len(clisteners) == 0:
					del self.listeners_by_channel[channel]
		finally:
			self.lock.release()

	def add_to_queues(self, channel, event):
		self.lock.acquire()
		try:
			wake = []
			listeners = self.listeners_by_channel.get(channel, set())
			for l in listeners:
				items = l.channel_items.get(channel)
				if items is None:
					items = []
					l.channel_items[channel] = items
				if len(items) < MAX_PENDING:
					items.append(event)
					wake.append(l)
				else:
					l.overflow = True
			for l in wake:
				l.wake_threadsafe()
		finally:
			self.lock.release()

	def kick(self, user_id, channel):
		self.lock.acquire()
		try:
			wake = []
			listeners = self.listeners_by_channel.get(channel, set())
			for l in listeners:
				if l.user_id == user_id:
					msg = 'Permission denied to channels: %s' % channel
					l.error = {'condition': 'forbidden', 'text': msg, 'channels': [channel]}
					wake.append(l)
			for l in wake:
				l.wake_threadsafe()
		finally:
			self.lock.release()

listener_manager = ListenerManager()

def get_listener_manager():
	return listener_manager

class EventsConsumer(AsyncHttpConsumer):
	# this function adapted from channels 2.1.5
	async def __call__(self, scope, receive, send):
		"""
		Async entrypoint - concatenates body fragments and hands off control
		to ``self.handle`` when the body has been completely received.
		"""
		self.scope = scope
		self.send = send
		body = []
		try:
			while True:
				message = await receive()
				if message["type"] == "http.disconnect":
					return
				else:
					if "body" in message:
						body.append(message["body"])
					if not message.get("more_body"):
						await self.handle(b"".join(body))
		finally:
			await self.disconnect()

	@database_sync_to_async
	def parse_request(self, request):
		from .eventrequest import EventRequest

		kwargs = self.scope['url_route']['kwargs']

		return EventRequest(request, view_kwargs=kwargs)

	@database_sync_to_async
	def get_events(self, event_request):
		from .eventstream import get_events

		return get_events(event_request)

	async def handle(self, body):
		from django_grip import GripMiddleware
		from .eventrequest import EventRequest
		from .eventstream import EventPermissionError
		from .utils import sse_error_response

		self.listener = None

		request = AsgiRequest(self.scope, body)

		def dummy_get_response(request):
			return None

		gm = GripMiddleware(dummy_get_response)
		gm.process_request(request)

		if 'user' in self.scope:
			request.user = self.scope['user']

		if 'session' in self.scope:
			request.session = self.scope['session']

		try:
			event_request = await self.parse_request(request)
			response = None
		except EventRequest.ResumeNotAllowedError as e:
			response = HttpResponseBadRequest(
				'Invalid request: %s.\n' % str(e))
		except EventRequest.GripError as e:
			if request.grip.proxied:
				response = sse_error_response(
					'internal-error',
					'Invalid internal request.')
			else:
				response = sse_error_response(
					'bad-request',
					'Invalid request: %s.' % str(e))
		except EventRequest.Error as e:
			response = sse_error_response(
				'bad-request',
				'Invalid request: %s.' % str(e))

		# for grip requests, prepare immediate response
		if not response and request.grip.proxied:
			try:
				event_response = await self.get_events(event_request)
				response = event_response.to_http_response(request)
			except EventPermissionError as e:
				response = sse_error_response(
					'forbidden',
					str(e),
					{'channels': e.channels})

		extra_headers = {}
		add_default_headers(extra_headers)

		# if this was a grip request or we encountered an error, respond now
		if response:
			response = gm.process_response(request, response)

			headers = []
			for name, value in response.items():
				if isinstance(name, six.text_type):
					name = name.encode('utf-8')
				if isinstance(value, six.text_type):
					value = value.encode('utf-8')
				headers.append((name, value))

			for name, value in extra_headers.items():
				if isinstance(name, six.text_type):
					name = name.encode('utf-8')
				if isinstance(value, six.text_type):
					value = value.encode('utf-8')
				headers.append((name, value))

			await self.send_response(
				response.status_code,
				response.content,
				headers=headers
			)
			return

		# if we got here then the request was not a grip request, and there
		#   were no errors, so we can begin a local stream response

		headers = [(six.b('Content-Type'), six.b('text/event-stream'))]
		for name, value in extra_headers.items():
			if isinstance(name, six.text_type):
				name = name.encode('utf-8')
			if isinstance(value, six.text_type):
				value = value.encode('utf-8')
			headers.append((name, value))

		await self.send_headers(headers=headers)

		self.listener = Listener()
		self.is_streaming = True

		asyncio.get_event_loop().create_task(self.stream(event_request))

	async def stream(self, event_request):
		from .eventstream import EventPermissionError
		from .utils import sse_encode_event, make_id

		self.listener.user_id = event_request.user.id if event_request.user else 'anonymous'
		self.listener.channels = event_request.channels

		lm = get_listener_manager()

		lm.add_listener(self.listener)

		first_result = True

		while self.is_streaming:
			try:
				event_response = await self.get_events(event_request)
			except EventPermissionError as e:
				data = {
					'condition': 'forbidden',
					'text': str(e),
					'channels': e.channels,
				}
				body = sse_encode_event('stream-error', data, event_id='error')
				await self.send_body(body.encode('utf-8'))
				break

			last_ids = copy.deepcopy(event_response.channel_last_ids)
			event_id = make_id(last_ids)

			body = ''

			if first_result:
				first_result = False

				# include padding on the first result
				body += ':' + (' ' * 2048) + '\n\n'
				body += 'event: stream-open\ndata:\n\n'

			if len(event_response.channel_reset) > 0:
				body += sse_encode_event(
					'stream-reset',
					{'channels': list(event_response.channel_reset)},
					event_id=event_id)

			for channel, items in event_response.channel_items.items():
				for item in items:
					last_ids[channel] = item.id
					event_id = make_id(last_ids)
					body += sse_encode_event(
						item.type,
						item.data,
						event_id=event_id)

			await self.send_body(body.encode('utf-8'), more_body=True)

			if len(event_response.channel_more) > 0:
				# read again immediately
				continue

			# FIXME: reconcile without re-reading from db

			lm.lock.acquire()
			conflict = False
			if len(self.listener.channel_items) > 0:
				# items were queued while reading from the db. toss them and
				#   read from db again
				self.listener.aevent.clear()
				self.listener.channel_items = {}
				conflict = True
			lm.lock.release()

			if conflict:
				continue

			# if we get here then the client is caught up. time to wait

			while self.is_streaming:
				f = asyncio.ensure_future(self.listener.aevent.wait())
				while True:
					done, _ = await asyncio.wait([f], timeout=20)
					if f in done:
						break
					body = 'event: keep-alive\ndata:\n\n'
					await self.send_body(body.encode('utf-8'), more_body=True)

				if not self.is_streaming:
					break

				lm.lock.acquire()

				channel_items = self.listener.channel_items
				overflow = self.listener.overflow
				error_data = self.listener.error

				self.listener.aevent.clear()
				self.listener.channel_items = {}
				self.listener.overflow = False

				lm.lock.release()

				body = ''
				for channel, items in channel_items.items():
					for item in items:
						if channel in last_ids:
							if item.id is not None:
								last_ids[channel] = item.id
							else:
								del last_ids[channel]
						if last_ids:
							event_id = make_id(last_ids)
						else:
							event_id = None
						body += sse_encode_event(
							item.type,
							item.data,
							event_id=event_id)

				more = True

				if error_data:
					body += sse_encode_event('stream-error', error_data, event_id='error')
					more = False

				if body or not more:
					await self.send_body(body.encode('utf-8'), more_body=more)

				if not more:
					self.is_streaming = False
					break

				if overflow:
					# check db
					break

			event_request.channel_last_ids = last_ids

		lm.remove_listener(self.listener)

	async def disconnect(self):
		self.is_streaming = False
		if self.listener:
			self.listener.aevent.set()



import time
import jwt
import six
from django.contrib.auth import get_user_model
from django.conf import settings
from .utils import parse_last_event_id, get_channelmanager

try:
	from urllib import unquote
except ImportError:
	from urllib.parse import unquote

class EventRequest(object):
	class Error(ValueError):
		pass

	class GripError(Error):
		pass

	class ResumeNotAllowedError(Error):
		pass

	def __init__(self, http_request=None, channel_limit=10, view_kwargs=None):
		if view_kwargs is None:
			view_kwargs = {}

		self.channels = set()
		self.channel_last_ids = {}
		self.is_recover = False
		self.user = None

		if http_request:
			self.apply_http_request(http_request,
				channel_limit=channel_limit,
				view_kwargs=view_kwargs)

	def apply_http_request(self, http_request, channel_limit, view_kwargs):
		is_next = False
		is_recover = False
		user = None

		es_meta = {}
		if http_request.GET.get('es-meta'):
			es_meta = jwt.decode(http_request.GET['es-meta'], settings.SECRET_KEY.encode('utf-8'), algorithms=['HS256'])
			if int(time.time()) >= es_meta['exp']:
				raise ValueError('es-meta signature is expired')

		if 'user' in es_meta:
			if es_meta['user'] != 'anonymous':
				user = get_user_model().objects.get(pk=es_meta['user'])
		else:
			if hasattr(http_request, 'user') and http_request.user.is_authenticated:
				user = http_request.user

		if 'channels' in es_meta:
			channels = es_meta['channels']
		else:
			channelmanager = get_channelmanager()
			channels = channelmanager.get_channels_for_request(http_request, view_kwargs)

		if len(channels) < 1:
			raise EventRequest.Error('No channels specified')

		if len(channels) > channel_limit:
			raise EventRequest.Error('Channel limit exceeded')

		if http_request.GET.get('link') == 'next':
			is_next = True

		if http_request.GET.get('recover') == 'true':
			channel_last_ids = {}
			is_recover = True
			for grip_channel, last_id in six.iteritems(http_request.grip.last):
				if not grip_channel.startswith('events-'):
					continue
				channel = unquote(grip_channel[7:])
				if channel in channels:
					channel_last_ids[channel] = last_id
		else:
			last_event_id = http_request.META.get('HTTP_LAST_EVENT_ID')
			if not last_event_id:
				# take the first non-empty param, from the end
				for val in reversed(http_request.GET.getlist('lastEventId')):
					if val:
						last_event_id = val
						break

			if last_event_id:
				if last_event_id == 'error':
					raise EventRequest.ResumeNotAllowedError(
						'Can\'t resume session after stream-error')

				try:
					parsed = parse_last_event_id(last_event_id)

					channel_last_ids = {}
					for channel, last_id in six.iteritems(parsed):
						channel = unquote(channel)
						if channel in channels:
							channel_last_ids[channel] = last_id
				except:
					raise EventRequest.Error(
						'Failed to parse Last-Event-ID or lastEventId')
			else:
				channel_last_ids = {}

		self.channels = channels
		self.channel_last_ids = channel_last_ids
		self.is_next = is_next
		self.is_recover = is_recover
		self.user = user



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.http import HttpResponseBadRequest
from .utils import add_default_headers

def events(request, **kwargs):
	from .eventrequest import EventRequest
	from .eventstream import EventPermissionError, get_events
	from .utils import sse_error_response

	try:
		event_request = EventRequest(request, view_kwargs=kwargs)
		event_response = get_events(event_request)
		response = event_response.to_http_response(request)
	except EventRequest.ResumeNotAllowedError as e:
		response = HttpResponseBadRequest(
			'Invalid request: %s.\n' % str(e))
	except EventRequest.GripError as e:
		if request.grip.proxied:
			response = sse_error_response(
				'internal-error',
				'Invalid internal request.')
		else:
			response = sse_error_response(
				'bad-request',
				'Invalid request: %s.' % str(e))
	except EventRequest.Error as e:
		response = sse_error_response(
			'bad-request',
			'Invalid request: %s.' % str(e))
	except EventPermissionError as e:
		response = sse_error_response(
			'forbidden',
			str(e),
			{'channels': e.channels})

	add_default_headers(response)

	return response



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
	readme = f.read()

install_requires = []

install_requires.extend(['PyJWT>=1.5,<3', 'gripcontrol>=4.0,<5', 'django_grip>=3.0,<4', 'six>=1.10,<2'])

setup(
name='django-eventstream',
version='4.4.0',
description='Server-Sent Events for Django',
long_description=readme,
long_description_content_type='text/markdown',
author='Justin Karneges',
author_email='justin@fanout.io',
url='https://github.com/fanout/django-eventstream',
license='MIT',
zip_safe=False,
packages=['django_eventstream', 'django_eventstream.migrations', 'django_eventstream.management', 'django_eventstream.management.commands'],
package_data={'django_eventstream': ['static/django_eventstream/*']},
install_requires=install_requires,
tests_require=['Django>=2.0'],
test_suite='tests.runtests.runtests',
classifiers=[
	'Development Status :: 4 - Beta',
	'Topic :: Utilities',
	'License :: OSI Approved :: MIT License',
	'Programming Language :: Python',
	'Programming Language :: Python :: 2',
	'Programming Language :: Python :: 3',
	'Framework :: Django',
]
)
